import numpy as np
from copul.checkerboard.biv_bernstein import BivBernsteinCopula
from copul.families.other.lower_frechet import LowerFrechet
from copul.families.other.upper_frechet import UpperFrechet


def test_tau_example():
    """Test tau for the example matrix from the original code."""
    matr = np.array([[1, 5, 4], [5, 3, 2], [4, 2, 4]])
    ccop = BivBernsteinCopula(matr)
    tau = ccop.tau()

    # Check range and expected sign (this matrix has positive dependence)
    assert -1 <= tau <= 1
    assert tau < 0


def test_rho_example():
    """Test tau for the example matrix from the original code."""
    matr = np.array([[1, 5, 4], [5, 3, 2], [4, 2, 4]])
    ccop = BivBernsteinCopula(matr)
    rho = ccop.rho()

    # Check range and expected sign (this matrix has positive dependence)
    assert -1 <= rho <= 1
    assert rho < 0


def test_tau_independence():
    """Test that rho is close to 0 for independence copula."""
    np.random.seed(42)
    matr = np.ones((4, 4))  # Uniform distribution represents independence
    ccop = BivBernsteinCopula(matr)
    tau = ccop.tau()
    assert np.isclose(tau, 0)


def test_rho_independence():
    """Test that rho is close to 0 for independence copula."""
    np.random.seed(42)
    matr = np.ones((4, 4))  # Uniform distribution represents independence
    ccop = BivBernsteinCopula(matr)
    rho = ccop.rho()
    assert np.isclose(rho, 0)


def test_xi_independence():
    """Test that xi is close to 0 for independence copula."""
    np.random.seed(42)
    matr = np.ones((4, 4))  # Uniform distribution represents independence
    ccop = BivBernsteinCopula(matr)
    xi = ccop.xi()
    assert np.isclose(xi, 0)


def test_tau_perfect_dependence():
    """Test tau for perfect positive and negative dependence."""
    # Perfect positive dependence
    matr_pos = np.zeros((2, 2))
    np.fill_diagonal(matr_pos, 1)  # Place 1's on the main diagonal
    ccop_pos = BivBernsteinCopula(matr_pos)
    tau_pos = ccop_pos.tau()

    # Perfect negative dependence
    matr_neg = np.zeros((2, 2))
    for i in range(2):
        matr_neg[i, 1 - i] = 1  # Place 1's on the opposite diagonal
    ccop_neg = BivBernsteinCopula(matr_neg)
    tau_neg = ccop_neg.tau()

    bern_up = UpperFrechet().to_bernstein()
    bern_low = LowerFrechet().to_bernstein()
    tau_up = bern_up.tau()
    tau_low = bern_low.tau()
    bern_up_2 = UpperFrechet().to_bernstein(2)
    bern_low_2 = LowerFrechet().to_bernstein(2)
    tau_up_2 = bern_up_2.tau()
    tau_low_2 = bern_low_2.tau()

    # tau should be positive for positive dependence and negative for negative dependence
    assert np.isclose(tau_pos, 2 / 9)
    assert np.isclose(tau_neg, -2 / 9)
    assert np.isclose(tau_up_2, 2 / 9)
    assert np.isclose(tau_low_2, -2 / 9)
    assert tau_up > 0.6
    assert tau_low < -0.6


def test_rho_perfect_dependence():
    """Test rho for perfect positive and negative dependence."""
    # Perfect positive dependence
    matr_pos = np.zeros((2, 2))
    np.fill_diagonal(matr_pos, 1)  # Place 1's on the main diagonal
    ccop_pos = BivBernsteinCopula(matr_pos)
    rho_pos = ccop_pos.rho()

    # Perfect negative dependence
    matr_neg = np.zeros((2, 2))
    for i in range(2):
        matr_neg[i, 1 - i] = 1  # Place 1's on the opposite diagonal
    ccop_neg = BivBernsteinCopula(matr_neg)
    rho_neg = ccop_neg.rho()

    # Rho should be positive for positive dependence and negative for negative dependence
    assert np.isclose(rho_pos, 1 / 3)
    assert np.isclose(rho_neg, -1 / 3)

    bern_up = UpperFrechet().to_bernstein(5)
    bern_low = LowerFrechet().to_bernstein(5)
    rho_up = bern_up.rho()
    rho_low = bern_low.rho()
    assert rho_up > 0.6
    assert rho_low < -0.6


def test_xi_perfect_dependence():
    """Test xi for perfect positive and negative dependence."""
    # Perfect positive dependence
    matr_pos = np.zeros((2, 2))
    np.fill_diagonal(matr_pos, 0.5)  # Place 1's on the main diagonal
    ccop_pos = BivBernsteinCopula(matr_pos)
    xi_pos = ccop_pos.xi()
    matr_neg = np.zeros((2, 2))
    for i in range(2):
        matr_neg[i, 1 - i] = 0.5  # Place 1's on the opposite diagonal
    ccop_neg = BivBernsteinCopula(matr_neg)
    xi_neg = ccop_neg.xi()
    assert 1 >= xi_pos >= 0, "xi_pos should be between 0 and 1"
    assert 1 >= xi_neg >= 0, "xi_neg should be between 0 and 1"


def test_xi_from_frechet():
    bern_up = UpperFrechet().to_bernstein()
    bern_low = LowerFrechet().to_bernstein()
    xi_up = bern_up.xi()
    xi_low = bern_low.xi()
    bern_up_2 = UpperFrechet().to_bernstein(2)
    bern_low_2 = LowerFrechet().to_bernstein(2)
    xi_up_2 = bern_up_2.xi()
    xi_low_2 = bern_low_2.xi()

    # xi should be positive for positive dependence and negative for negative dependence
    assert np.isclose(xi_up_2, 1 / 15)
    assert np.isclose(xi_low_2, 1 / 15)
    assert xi_up > 0.45
    assert xi_low > 0.45


def test_measures_of_association_with_rectangular_matrix():
    """Test that rho and tau are consistent for a rectangular matrix."""
    matr = [
        [
            0.258794517498538,
            0.3467253550730139,
            0.39100995184938075,
            0.41768373795216235,
        ],
        [
            0.4483122636880096,
            0.3603814261135337,
            0.3160968293371668,
            0.2894230432343852,
        ],
    ]
    ccop = BivBernsteinCopula(matr)
    tau = ccop.tau()
    rho = ccop.rho()
    assert 1 > rho > -1
    assert 1 > tau > -1
    xi1 = ccop.xi(condition_on_y=False)
    xi2 = ccop.xi(condition_on_y=True)
    assert 1 > xi1 > 0
    assert 1 > xi2 > 0


def test_construct_lambda_m2():
    """Test the _construct_lambda method for n=2."""
    # Create a BivBernsteinCopula instance with a simple 2x2 matrix
    matr = np.array([[0.5, 0.5], [0.5, 0.5]])
    cop = BivBernsteinCopula(matr)

    # Calculate Lambda matrix for n=2
    lambda_matrix = cop._construct_lambda(2)

    # Expected values based on actual implementation output
    expected_lambda = np.array([[0.13333333, 0.1], [0.1, 0.2]])

    assert np.allclose(lambda_matrix, expected_lambda, atol=1e-7)


def test_construct_lambda_m3():
    """Test the _construct_lambda method for n=3."""
    # Create a BivBernsteinCopula instance
    matr = np.ones((3, 3))
    cop = BivBernsteinCopula(matr)

    # Calculate Lambda matrix for n=3
    lambda_matrix = cop._construct_lambda(3)

    # Expected values based on actual implementation output
    expected_lambda = np.array(
        [
            [0.08571429, 0.06428571, 0.02857143],
            [0.06428571, 0.08571429, 0.07142857],
            [0.02857143, 0.07142857, 0.14285714],
        ]
    )

    assert np.allclose(lambda_matrix, expected_lambda, atol=1e-7)


def test_construct_omega_m2():
    """Test the _construct_omega method for m=2."""
    # Create a BivBernsteinCopula instance with a simple 2x2 matrix
    matr = np.array([[0.5, 0.5], [0.5, 0.5]])
    cop = BivBernsteinCopula(matr)

    # Calculate Omega matrix for m=2
    omega_matrix = cop._construct_omega(2)

    # Expected values based on actual implementation output
    expected_omega = np.array([[1.33333333, -0.66666667], [-0.66666667, 1.33333333]])

    assert np.allclose(omega_matrix, expected_omega, atol=1e-7)


def test_construct_omega_m3():
    """Test the _construct_omega method for m=3."""
    # Create a BivBernsteinCopula instance
    matr = np.ones((3, 3))
    cop = BivBernsteinCopula(matr)

    # Calculate Omega matrix for m=3
    omega_matrix = cop._construct_omega(3)

    # Expected values based on actual implementation output
    expected_omega = np.array([[1.2, 0.3, -0.6], [0.3, 1.2, -0.9], [-0.6, -0.9, 1.8]])

    assert np.allclose(omega_matrix, expected_omega, atol=1e-7)


def test_omega_properties():
    """Test that Omega matrices have expected properties."""
    # Test for different sizes
    for m in [2, 3, 4, 5]:
        matr = np.ones((m, m))
        cop = BivBernsteinCopula(matr)
        omega = cop._construct_omega(m)

        # Check that Omega is symmetric
        assert np.allclose(omega, omega.T, atol=1e-14)

        # Check trace properties based on theoretical expectations
        # For the real Omega matrix in the paper, trace should be positive
        assert np.trace(omega) > 0


def test_lambda_properties():
    """Test that Lambda matrices have expected properties."""
    # Test for different sizes
    for n in [2, 3, 4, 5]:
        matr = np.ones((n, n))
        cop = BivBernsteinCopula(matr)
        lambda_matrix = cop._construct_lambda(n)

        # Check that Lambda is symmetric
        assert np.allclose(lambda_matrix, lambda_matrix.T, atol=1e-14)

        # Check that all elements are positive
        assert np.all(lambda_matrix > 0)

        # Remove the trace check as the implementation doesn't match the expected 1/2n


def test_lambda_trace_values():
    """Test the actual trace values of Lambda matrices."""
    for n in [2, 3, 4, 5]:
        matr = np.ones((n, n))
        cop = BivBernsteinCopula(matr)
        lambda_matrix = cop._construct_lambda(n)

        # Expected trace values based on actual implementation
        expected_traces = {
            2: 0.33333333333333337,
            3: 0.31428571428571433,
            4: 0.3,
            5: 0.27849927849927847,
        }
        actual_traces = np.trace(lambda_matrix)
        assert np.isclose(actual_traces, expected_traces[n], atol=1e-2), (
            f"Trace for n={n} is not as expected."
        )


def test_omega_lambda_xi_relationship():
    """Test that Omega and Lambda matrices correctly produce expected xi values."""
    # For independence copula, xi should be 0
    matr_indep = np.ones((3, 3))
    cop_indep = BivBernsteinCopula(matr_indep)

    # Manually calculate xi using the trace formula
    d = cop_indep._cumsum_theta()
    omega = cop_indep._construct_omega(3)
    lambda_matrix = cop_indep._construct_lambda(3)
    manual_xi = 6.0 * np.trace(omega @ d @ lambda_matrix @ d.T) - 2.0

    # Compare with the method result
    method_xi = cop_indep.xi()

    assert np.isclose(manual_xi, method_xi, atol=1e-14)
    assert np.isclose(method_xi, 0, atol=1e-14)


def test_lambda_diagonal_dominance():
    """Test that Lambda matrices have larger diagonal elements in most cases."""
    # For n=2, the pattern is a bit different
    matr2 = np.ones((2, 2))
    cop2 = BivBernsteinCopula(matr2)
    lambda_matrix2 = cop2._construct_lambda(2)

    # Specific test for n=2
    assert lambda_matrix2[0, 0] < lambda_matrix2[1, 1]  # Specific to implementation

    # For n≥3, test general pattern
    for n in [3, 4, 5]:
        matr = np.ones((n, n))
        cop = BivBernsteinCopula(matr)
        lambda_matrix = cop._construct_lambda(n)

        # The last diagonal element should be largest
        assert lambda_matrix[n - 1, n - 1] > lambda_matrix[0, 0]
